import re
import sys

import pytz
from dateutil.parser import parse

from cert_core import *

V1_1_REGEX = re.compile('[0-9a-fA-F]{24}')
V1_2_REGEX = re.compile('[0-9a-fA-F]{8}\-[0-9a-fA-F]{4}\-[0-9a-fA-F]{4}\-[0-9a-fA-F]{4}\-[0-9a-fA-F]{12}')

CONTEXT_REGEX = re.compile('v[0-9]')
USE_PREFIX = False

URN_UUID_PREFIX = 'urn:uuid:'


def scope_name(name):
    """
    TBD whether we want to include prefix. Doing this for now. Default is no prefix
    :param name:
    :return:
    """
    if USE_PREFIX:
        return BLOCKCERTS_PREFIX + name
    else:
        return name


class ProofType(Enum):
    chainpoint_sha256_v2 = 0,
    pop_2016 = 1


class HashProof(object):
    def __init__(self, transaction_id):
        self.transaction_id = transaction_id


class MerkleProof(HashProof):
    def __init__(self, transaction_id, target_hash, merkle_root, proof_type, proof_json):
        super(MerkleProof, self).__init__(transaction_id)
        self.target_hash = target_hash
        self.merkle_root = merkle_root
        self.proof_type = proof_type
        self.proof_json = proof_json


class SignatureLine(object):
    def __init__(self, image, name=None, job_title=None):
        self.image = image
        self.name = name
        self.job_title = job_title


class BlockcertSignature(object):
    def __init__(self, type, creator, created, signature_value, merkle_proof):
        self.type = type
        self.creator = creator
        self.created = created
        self.signature_value = signature_value
        self.merkle_proof = merkle_proof


class SpendToRevokeBlockcertSignature(BlockcertSignature):
    def __init__(self, type, creator, created, signature_value, merkle_proof, recipient_public_key,
                 per_recipient_revocation_key):
        super(SpendToRevokeBlockcertSignature, self).__init__(type, creator, created, signature_value, merkle_proof)
        self.recipient_public_key = recipient_public_key
        self.per_recipient_revocation_key = per_recipient_revocation_key


class BlockchainCertificate(object):
    def __init__(self, version, uid, logo_image, recipient_name, recipient_public_key, title,
                 organization, description, signature_image, issuer_id, issued_on, expires, subtitle,
                 blockcert_signature, certificate_json, document):
        self.version = version
        self.uid = uid
        self.logo_image = logo_image
        self.recipient_name = recipient_name
        self.recipient_public_key = recipient_public_key
        self.title = title
        self.organization = organization
        self.description = description
        self.signature_image = signature_image
        self.issuer_id = issuer_id
        self.issued_on = issued_on
        self.expires = expires
        self.subtitle = subtitle
        self.blockcert_signature = blockcert_signature
        self.certificate_json = certificate_json
        self.document = document

    def get_transaction_id(self):
        return self.blockcert_signature.merkle_proof.transaction_id

    def get_signature_value(self):
        return self.blockcert_signature.signature_value

    def __str__(self):
        sb = []
        for key in self.__dict__:
            _value = self.__dict__[key]

            if _value and isstring(_value) and 'data:image/png;base64' in str(_value):
                mapped_value = '<base64_encoded_image>'
            elif _value:
                mapped_value = _value
            else:
                mapped_value = '<None>'
            sb.append("{key}='{value}'".format(key=key, value=mapped_value))

        return ', '.join(sb)

    def __repr__(self):
        return self.__str__()


def isstring(s):
    if (sys.version_info[0] >= 3):
        return isinstance(s, str)
    return isinstance(s, basestring)


def detect_version(certificate_json):
    # assumes it's a certificate. Maybe add some schema validation
    if not '@context' in certificate_json:
        return BlockcertVersion.V1_1
    context = certificate_json['@context']
    if isinstance(context, list):
        version_marker = context[0]
    else:
        version_marker = context

    matches = CONTEXT_REGEX.search(version_marker)
    if matches:
        version = matches.group(0)
        if version == 'v1':
            return BlockcertVersion.V1_2
        elif version == 'v2':
            return BlockcertVersion.V2

    raise UnknownBlockcertVersionException()


def is_v1_uid(uid):
    if V1_1_REGEX.search(uid):
        return True
    else:
        return False


def parse_chainpoint_proof(receipt):
    proof_type = parse_proof_type(receipt['type'])
    return MerkleProof(receipt['anchors'][0]['sourceId'], receipt['targetHash'], receipt['merkleRoot'],
                       proof_type, receipt)


def parse_proof_type(value):
    if value == 'ChainpointSHA256v2':
        return ProofType.chainpoint_sha256_v2
    else:
        return None


def parse_signature(signature_json, issuer_id, issued_on):
    return BlockcertSignature(signature_json['type'], issuer_id, issued_on,
                              signature_json['signatureValue'], parse_chainpoint_proof(signature_json['merkleProof']))


def parse_date(raw_date):
    if raw_date is None:
        return None
    parsed_date = parse(raw_date)
    utc = pytz.UTC
    if parsed_date.tzinfo is None or parsed_date.tzinfo.utcoffset(parsed_date) is None:
        parsed_date = utc.localize(parsed_date)
    return parsed_date


def parse_expires_date(assertion):
    if 'expires' in assertion:
        return parse_date(assertion['expires'])
    else:
        return None


def parse_v2_blockchain_certificate(certificate_json):
    assertion = certificate_json
    uid = assertion['id']
    badge = assertion['badge']
    recipient = assertion['recipient']
    issuer_id = badge['issuer']['id']
    issued_on = parse_date(assertion['issuedOn'])
    signature = assertion[scope_name('signature')]
    blockcert_signature = parse_signature(signature, issuer_id, issued_on)

    signature_lines = []
    if scope_name('signatureLines') in badge:
        signature_lines_raw = badge[scope_name('signatureLines')]
        for l in signature_lines_raw:
            signature_lines.append(SignatureLine(l['image'], l['name'], l['jobTitle']))

    if 'subtitle' in badge:
        subtitle = badge['subtitle']
    else:
        subtitle = None

    recipient_public_key_full = recipient[scope_name('recipientProfile')]['publicKey']

    recipient_public_key = str.split(str(recipient_public_key_full), ':')[1]

    import copy
    document_json = copy.deepcopy(certificate_json)
    del document_json['signature']

    return BlockchainCertificate(BlockcertVersion.V2,
                                 uid,
                                 badge['issuer']['image'],
                                 recipient[scope_name('recipientProfile')]['name'],
                                 recipient_public_key,
                                 badge['name'],
                                 badge['issuer']['name'],
                                 badge['description'],
                                 signature_lines,
                                 issuer_id,
                                 issued_on,
                                 parse_expires_date(assertion),
                                 subtitle,
                                 blockcert_signature,
                                 certificate_json,
                                 document_json)


def parse_v1_2_blockchain_certificate(certificate_json):
    document = certificate_json['document']
    certificate = document['certificate']
    assertion = document['assertion']
    recipient = document['recipient']
    recipient_public_key = recipient['publicKey']
    issued_on = parse_date(assertion['issuedOn'])
    issuer_id = certificate['issuer']['id']
    signature_lines = []
    if 'image:signature' in assertion:
        signature_lines.append(SignatureLine(assertion['image:signature']))
    if 'subtitle' in certificate:
        subtitle = certificate['subtitle']
    else:
        subtitle = None

    if 'revocationKey' in recipient:
        recipient_revocation_address = recipient['revocationKey']
    else:
        recipient_revocation_address = None

    blockcert_signature = SpendToRevokeBlockcertSignature('EcdsaKoblitzSignature2016',
                                                          issuer_id,
                                                          issued_on,
                                                          document['signature'],
                                                          parse_chainpoint_proof(certificate_json['receipt']),
                                                          recipient_public_key,
                                                          recipient_revocation_address)

    return BlockchainCertificate(BlockcertVersion.V1_2,
                                 assertion['uid'],
                                 certificate['issuer']['image'],
                                 recipient['givenName'] + ' ' + recipient['familyName'],
                                 recipient_public_key,
                                 certificate['name'],
                                 certificate['issuer']['name'],
                                 certificate['description'],
                                 signature_lines,
                                 issuer_id,
                                 issued_on,
                                 parse_expires_date(assertion),
                                 subtitle,
                                 blockcert_signature,
                                 certificate_json,
                                 document)


def parse_v1_1_blockchain_certificate(json_certificate, txid, certificate_bytes):
    subtitle = json_certificate['certificate']['subtitle']['content']
    display_subtitle = json_certificate['certificate']['subtitle']['display']
    if display_subtitle in ['true', 'True', 'TRUE']:
        subtitle = subtitle
    else:
        subtitle = None

    issuer_id = json_certificate['certificate']['issuer']['id']
    issued_on = parse_date(json_certificate['assertion']['issuedOn'])
    recipient_pubkey = json_certificate['recipient']['pubkey']
    blockcert_signature = SpendToRevokeBlockcertSignature('EcdsaKoblitzSignature2016', issuer_id, issued_on,
                                                          json_certificate['signature'],
                                                          HashProof(txid),
                                                          recipient_pubkey, None)
    signature_lines = []
    if 'image:signature' in json_certificate['assertion']:
        signature_lines.append(SignatureLine(json_certificate['assertion']['image:signature']))
    return BlockchainCertificate(BlockcertVersion.V1_1,
                                 json_certificate['assertion']['uid'],
                                 json_certificate['certificate']['issuer']['image'],
                                 json_certificate['recipient']['givenName'] + ' ' + json_certificate['recipient'][
                                     'familyName'],
                                 recipient_pubkey,
                                 json_certificate['certificate']['title'],
                                 json_certificate['certificate']['issuer']['name'],
                                 json_certificate['certificate']['description'],
                                 signature_lines,
                                 issuer_id,
                                 issued_on,
                                 parse_expires_date(json_certificate['assertion']),
                                 subtitle,
                                 blockcert_signature,
                                 json_certificate,
                                 certificate_bytes)


def to_certificate_model(certificate_json, txid=None, certificate_bytes=None):
    version = detect_version(certificate_json)
    if version == BlockcertVersion.V1_1:
        if not txid or not certificate_bytes:
            raise Exception('V1.1 Blockchain Certificates require a transaction id and raw bytes')
        return parse_v1_1_blockchain_certificate(certificate_json, txid, certificate_bytes)
    elif version == BlockcertVersion.V1_2:
        return parse_v1_2_blockchain_certificate(certificate_json)
    elif version == BlockcertVersion.V2:
        return parse_v2_blockchain_certificate(certificate_json)
    else:
        raise UnknownBlockcertVersionException(version)
