import json
import unittest

from cert_core import BlockcertVersion
from cert_core import helpers
from cert_core import model, Chain
from cert_core.model import ProofType, parse_date


class TestModel(unittest.TestCase):
    def test_detect_version_v1_1(self):
        with open('data/1.1/sample_signed_cert-1.1.json', 'rb') as cert_file:
            certificate_bytes = cert_file.read()
            certificate_json = helpers.certificate_bytes_to_json(certificate_bytes)
            version = model.detect_version(certificate_json)
            self.assertEquals(version, BlockcertVersion.V1_1)

    def test_detect_version_v1_2(self):
        with open('data/1.2/609c2989-275f-4f4c-ab02-b245cfb09017.json', 'rb') as cert_file:
            certificate_bytes = cert_file.read()
            certificate_json = helpers.certificate_bytes_to_json(certificate_bytes)
            version = model.detect_version(certificate_json)
            self.assertEquals(version, BlockcertVersion.V1_2)

    def test_to_certificate_model_v1_1(self):
        with open('data/1.1/sample_signed_cert-1.1.json', 'rb') as cert_file:
            certificate_bytes = cert_file.read()
            certificate_json = helpers.certificate_bytes_to_json(certificate_bytes)
            txid = '1703d2f5d706d495c1c65b40a086991ab755cc0a02bef51cd4aff9ed7a8586aa'
            certificate_model = model.to_certificate_model(certificate_json, txid, certificate_bytes)
            self.assertEquals(certificate_model.issuer_id,
                              'http://www.blockcerts.org/mockissuer/issuer/got-issuer.json')
            self.assertEquals(certificate_model.get_transaction_id(),
                              '1703d2f5d706d495c1c65b40a086991ab755cc0a02bef51cd4aff9ed7a8586aa')
            self.assertEquals(certificate_model.title, 'Game of Thrones Character')
            self.assertEquals(certificate_model.description,
                              'This certifies that the named character is an official Game of Thrones character.')
            self.assertEquals(certificate_model.uid, '75857d18-0e1c-4933-b4c8-33484396e06b')
            self.assertEquals(certificate_model.subtitle, None)
            self.assertEquals(certificate_model.expires, None)

            self.assertEquals(certificate_model.organization, 'Game of thrones characters')
            self.assertEquals(certificate_model.recipient_public_key, 'mgCNaPM3TFhh8Yn6U6VcEM9jWLhQbizy1x')
            self.assertEquals(certificate_model.issued_on, parse_date('2016-08-29 00:00:00+00:00'))
            # self.assertEquals(certificate_model.blockcert_signature, None)
            self.assertIsNotNone(certificate_model.signature_image[0].image)

    def test_to_certificate_model_v1_2(self):
        """
        Note this is a mainnet certificate with different uid
        :return:
        """
        with open('data/1.2/609c2989-275f-4f4c-ab02-b245cfb09017.json', 'rb') as cert_file:
            certificate_bytes = cert_file.read()
            certificate_json = helpers.certificate_bytes_to_json(certificate_bytes)
            certificate_model = model.to_certificate_model(certificate_json)
            self.assertEquals(certificate_model.version, BlockcertVersion.V1_2)
            self.assertEquals(certificate_model.issuer_id,
                              'http://www.blockcerts.org/mockissuer/issuer/got-issuer_live.json')
            self.assertEquals(certificate_model.get_transaction_id(),
                              '8623beadbc7877a9e20fb7f83eda6c1a1fc350171f0714ff6c6c4054018eb54d')
            self.assertEquals(certificate_model.title, 'Game of Thrones Character')
            self.assertEquals(certificate_model.description,
                              'This certifies that the named character is an official Game of Thrones character.')
            self.assertEquals(certificate_model.uid, '609c2989-275f-4f4c-ab02-b245cfb09017')
            self.assertEquals(certificate_model.subtitle, None)
            self.assertEquals(certificate_model.expires, None)

            self.assertEquals(certificate_model.organization, 'Game of thrones issuer')
            self.assertEquals(certificate_model.recipient_public_key, '1AAGG6jirbu9XwikFpkHokbbiYpjVtFe1G')
            self.assertEquals(certificate_model.issued_on, parse_date('2016-10-03 00:00:00+00:00'))
            self.assertIsNotNone(certificate_model.signature_image[0].image)

    def test_to_certificate_model_v2(self):
        with open('data/2.0/1dabb524-8365-4550-b499-a8119f6bf5d2.json', 'rb') as cert_file:
            certificate_bytes = cert_file.read()
            certificate_json = helpers.certificate_bytes_to_json(certificate_bytes)
            certificate_model = model.to_certificate_model(certificate_json)
            self.assertEquals(certificate_model.version, BlockcertVersion.V2)
            self.assertEquals(certificate_model.issuer_id, 'https://www.blockcerts.org/blockcerts_v2_alpha/samples/issuer_testnet.json')
            self.assertEquals(certificate_model.get_transaction_id(),
                              'e7a5441a231cb55327795e87af0568dc6e15c68bf1d805d385350c4af006c50d')
            self.assertEquals(certificate_model.title, 'This is the certificate title')
            self.assertEquals(certificate_model.description, 'This is the display description of the certificate.')
            self.assertEquals(certificate_model.uid, 'urn:uuid:1dabb524-8365-4550-b499-a8119f6bf5d2')
            self.assertEquals(certificate_model.subtitle, None)
            self.assertEquals(certificate_model.expires, None)

            self.assertEquals(certificate_model.organization, 'Issuer Institution Name')
            self.assertEquals(certificate_model.recipient_public_key, '1AAGG6jirbu9XwikFpkHokbbiYpjVtFe1G')
            self.assertEquals(certificate_model.issued_on, parse_date('2017-04-03 00:00:00+00:00'))

    def test_to_proof(self):
        with open('data/1.2/receipt.json') as receipt_file:
            receipt_json = json.load(receipt_file)
            proof = model.parse_chainpoint_proof(receipt_json['receipt'])
            self.assertEquals(proof.target_hash, '2d3d0f49416587c4ce14c05d47c4193f0da3dd56f7244a568b06484cb8d2fe78')
            self.assertEquals(proof.merkle_root, '3268386feb897f6cab2c100a0edb6f66b4bb3a8745e3e3e8a54b1cb7151a6d96')
            self.assertEquals(proof.proof_type, ProofType.chainpoint_sha256_v2)
            self.assertEquals(proof.transaction_id,
                              '7d0edf184a2969a9df3a129a9bd36e4b5f52b5faee7613c2b26691328ef62a56')

    def test_to_v2_proof(self):
        with open('data/2.0/receipt.json') as receipt_file:
            receipt_json = json.load(receipt_file)
            signature = model.parse_signature(receipt_json, '1', '2')
            self.assertEquals(signature.merkle_proof.target_hash,
                              'c9ead76a54426b4ce4899bb921e48f5b55ea7592e5cee4460c86ebf4698ac3a6')
            self.assertEquals(signature.merkle_proof.merkle_root,
                              '68f3ede17fdb67ffd4a5164b5687a71f9fbb68da803b803935720f2aa38f7728')
            self.assertEquals(signature.merkle_proof.proof_type, ProofType.chainpoint_sha256_v2)
            self.assertEquals(signature.merkle_proof.transaction_id,
                              '8623beadbc7877a9e20fb7f83eda6c1a1fc350171f0714ff6c6c4054018eb54d')


if __name__ == '__main__':
    unittest.main()
